import 'dart:collection';

import 'package:dio/dio.dart';

///
/// @author vay
/// @email hxhvay@126.com
/// @date 2020-05-09
///

class NetCache extends Interceptor {
  // 为确保迭代器顺序和对象插入时间一致顺序一致，我们使用LinkedHashMap
  var _cache = LinkedHashMap<String, CacheObject>();

  /// 是否启用缓存
  bool enable;

  /// 最大缓存数
  int maxCount;

  /// 缓存的最长时间，单位（秒）
  int maxAge;

  NetCache({this.enable = true, this.maxCount = 100, this.maxAge = 3600});

  @override
  onRequest(RequestOptions options) async {
    if (!this.enable) return options;
    // refresh标记是否是"下拉刷新"
    bool refresh = options.extra["refresh"] == true;
    // 如果是下拉刷新，先删除相关缓存
    if (refresh) {
      if (options.extra["list"] == true) {
        // 若是列表，则只要url中包含当前path的缓存全部删除（简单实现，并不精准）
        _cache.removeWhere((key, v) => key.contains(options.path));
      } else {
        // 如果不是列表，则只删除uri相同的缓存
        delete(options.uri.toString());
      }
      return options;
    }
    if (options.extra["noCache"] != true &&
        options.method.toLowerCase() == 'get') {
      String key = options.extra["cacheKey"] ?? options.uri.toString();
      var ob = _cache[key];
      if (ob != null) {
        // 若缓存未过期，则返回缓存内容
        if ((DateTime.now().millisecondsSinceEpoch - ob.timeStamp) / 1000 <
            this.maxAge) {
          return _cache[key].response;
        } else {
          // 若已过期则删除缓存，继续向服务器请求
          _cache.remove(key);
        }
      }
    }
  }

  @override
  onError(DioError err) async {
    // 错误状态不缓存
  }

  @override
  onResponse(Response response) async {
    // 如果启用缓存，将返回结果保存到缓存
    if (this.enable) {
      _saveCache(response);
    }
  }

  _saveCache(Response object) {
    RequestOptions options = object.request;
    if (options.extra["noCache"] != true &&
        options.method.toLowerCase() == "get") {
      // 如果缓存数量超过最大数量限制，则先移除最早的一条记录
      if (_cache.length == this.maxCount) {
        //_cache.remove(_cache[_cache.keys.first]);
        _cache.remove(_cache.keys.first);
      }
      String key = options.extra["cacheKey"] ?? options.uri.toString();
      _cache[key] = CacheObject(object);
    }
  }

  void delete(String key) {
    _cache.remove(key);
  }
}

class CacheObject {
  CacheObject(this.response)
      : timeStamp = DateTime.now().millisecondsSinceEpoch;
  Response response;

  /// 缓存创建时间
  int timeStamp;

  @override
  bool operator ==(other) {
    return response.hashCode == other.hashCode;
  }

  // 将请求uri作为缓存的key
  @override
  int get hashCode => response.realUri.hashCode;
}
